'*****************************************************************************
'                                                                            *
'  Columns.bas                                                               *
'                                                                            *
'  This program draws a grid of columns in 3D world space. It demonstrates   *
'  polygon culling and Fastgraph's incremental POV functions.                *
'                                                                            *
'*****************************************************************************

#COMPILE EXE "Columns.exe"

#INCLUDE "FGWin.inc"

%vbWidth = 600
%vbHeight = 400

%InfoHeight = 80

%WinWidth = %vbWidth
%WinHeight = (%vbHeight + %InfoHeight)

' six faces of a 2x2x10 column, defined in object coordinates
GLOBAL ColumnData() AS DOUBLE

'------------------------------------------------------------------------------

FUNCTION WINMAIN (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  lpCmdLine           AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

  LOCAL Msg         AS tagMsg
  LOCAL wndclass    AS WndClassEx
  LOCAL szClassName AS ASCIIZ * 80
  LOCAL hWnd        AS LONG

  DIM ColumnData(1:12,1:6)

  szClassName            = "FGcolumns"

  wndclass.cbSize        = SIZEOF(WndClass)
  wndclass.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_OWNDC
  wndclass.lpfnWndProc   = CODEPTR(WindowProc)
  wndclass.cbClsExtra    = 0
  wndclass.cbWndExtra    = 0
  wndclass.hInstance     = hInstance
  wndclass.hIcon         = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  wndclass.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wndclass.hbrBackground = %NULL
  wndclass.lpszMenuName  = %NULL
  wndclass.lpszClassName = VARPTR(szClassName)
  wndclass.hIconSm       = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  RegisterClassEx wndclass

  hWnd = CreateWindow( _
    szClassName, _                    ' window class name
    "3D Columns", _                   ' window caption
    %WS_POPUPWINDOW OR %WS_CAPTION, _ ' window style
    0, _                              ' initial x position
    0, _                              ' initial y position
    %WinWidth, _                      ' initial x size
    %WinHeight, _                     ' initial y size
    %NULL, _                          ' parent window handle
    %NULL, _                          ' window menu handle
    hInstance, _                      ' program instance handle
    BYVAL %NULL)                      ' creation parameters

  ShowWindow hWnd, iCmdShow
  UpdateWindow hWnd

  WHILE %TRUE
    IF PeekMessage(Msg, %NULL, 0, 0, %PM_REMOVE) THEN
      IF Msg.message = %WM_QUIT THEN
        EXIT LOOP
      ELSE
        TranslateMessage Msg
        DispatchMessage Msg
      END IF
    ELSE
      CALL CheckForMotion
    END IF
  WEND

  FUNCTION = Msg.wParam

END FUNCTION

'------------------------------------------------------------------------------

GLOBAL hDC AS LONG
GLOBAL hPal AS LONG
GLOBAL hVB AS LONG
GLOBAL hZB AS LONG

FUNCTION WindowProc (BYVAL hWnd AS LONG, BYVAL iMsg AS LONG, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) EXPORT AS LONG

  LOCAL ps AS PAINTSTRUCT

  SELECT CASE iMsg

    CASE %WM_CREATE
      ' set up the device context and logical palette
      hDC = GetDC(hWnd)
      fg_setdc hDC
      hPal = fg_defpal()
      fg_realize hPal

      ' initialize the virtual buffer environment
      fg_vbinit
      fg_vbdepth fg_colors()

      ' create and open the virtual buffer
      hVB = fg_vballoc(%vbWidth, %vbHeight)
      fg_vbopen hVB
      fg_vbcolors

      ' create and open the z-buffer
      hZB = fg_zballoc(%vbWidth, %vbHeight)
      fg_zbopen hZB

      ' define 3D viewport, render state, and initial POV
      fg_3Dviewport 0, %vbWidth-1, 0, %vbHeight-1, 1.0
      fg_3Drenderstate %FG_ZBUFFER + %FG_ZCLIP
      fg_3Dlookat 10.0, 20.0, 100.0, 0.0, 20.0, 0.0

      ' direct strings to the active virtual buffer
      fg_fontdc fg_getdc()

      ' make the client area equal to the required size
      ShowWindow hWnd, %SW_SHOWNORMAL
      CALL SetWindowSize(%WinWidth, %WinHeight)

      CALL InitGlobals

      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_PAINT
      BeginPaint hWnd, ps
      CALL DrawColumns
      EndPaint hWnd, ps
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SETFOCUS
      fg_realize hPal
      InvalidateRect hWnd, BYVAL %NULL, %TRUE
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_DESTROY
      fg_vbclose
      fg_zbfree hZB
      fg_vbfree hVB
      fg_vbfin
      DeleteObject hPal
      ReleaseDC hWnd, hDC
      PostQuitMessage 0
      FUNCTION = 0
      EXIT FUNCTION

  END SELECT

  FUNCTION = DefWindowProc(hWnd, iMsg, wParam, lParam)

END FUNCTION

'*****************************************************************************
'                                                                            *
'  CheckForMotion()                                                          *
'                                                                            *
'  The CheckForMotion() function checks for key presses that control the     *
'  viewer's position and orientation, and if required redraws the scene at   *
'  its new POV. It is called from the WinMain() message loop when there are  *
'  no messages waiting.                                                      *
'                                                                            *
'*****************************************************************************

SUB CheckForMotion()

  LOCAL ShiftKey AS LONG

  ' check if either shift key is pressed
  ShiftKey = (fg_kbtest(42) = 1) OR (fg_kbtest(54) = 1)

  IF fg_kbtest(71) = 1 THEN      ' Home
    CALL fg_3Dmoveup(5.0)
    CALL DrawColumns

  ELSEIF fg_kbtest(72) = 1 THEN  ' Up arrow
    CALL fg_3Dmoveforward(5.0)
    CALL DrawColumns

  ELSEIF fg_kbtest(73) = 1 THEN  ' PgUp
    CALL fg_3Drotateup(100)
    CALL DrawColumns

  ELSEIF fg_kbtest(75) = 1 THEN  ' Left arrow
    IF (ShiftKey) THEN
      CALL fg_3Dmoveright(-5.0)
    ELSE
      CALL fg_3Drotateright(-100)
    END IF
    CALL DrawColumns

  ELSEIF fg_kbtest(77) = 1 THEN  ' Right arrow
    IF (ShiftKey) THEN
      CALL fg_3Dmoveright(5.0)
    ELSE
      CALL fg_3Drotateright(100)
    END IF
    CALL DrawColumns

  ELSEIF fg_kbtest(79) = 1 THEN  ' End
    CALL fg_3Dmoveup(-5.0)
    CALL DrawColumns

  ELSEIF fg_kbtest(80) = 1 THEN  ' Down arrow
    CALL fg_3Dmoveforward(-5.0)
    CALL DrawColumns

  ELSEIF fg_kbtest(81) = 1 THEN  ' PgDn
    CALL fg_3Drotateup(-100)
    CALL DrawColumns
  END IF

END SUB

'*****************************************************************************
'                                                                            *
'  DrawColumns()                                                             *
'                                                                            *
'  Draws the scene at its new POV. Columns behind the viewer are culled out. *
'                                                                            *
'*****************************************************************************

SUB DrawColumns()

  DIM nColor(1:6) AS LONG
  LOCAL Row AS LONG, Col AS LONG
  LOCAL I AS LONG

  ' prepare for the new frame
  fg_zbframe
  fg_setcolor -1
  fg_fillpage

  ' create the six encoded color values
  nColor(1) = fg_maprgb(254,219,164)
  nColor(2) = fg_maprgb(243,194,117)
  nColor(3) = fg_maprgb(226,172, 86)
  nColor(4) = fg_maprgb(203,150, 67)
  nColor(5) = fg_maprgb(123, 98, 59)
  nColor(6) = fg_maprgb(166,125, 60)

  ' 50x50x6 = 15000 polygons per frame
  FOR Row = -500 TO 480 STEP 20
    FOR Col = -500 TO 480 STEP 20
      IF fg_3Dbehindviewer(CDBL(Row), 0.0, CDBL(Col), -1.0) = 0 THEN
        fg_3Dmoveobject CDBL(Row), 0.0, CDBL(Col)

        ' draw all the faces
        FOR I = 1 TO 6
          ' set the color
          fg_setcolor nColor(I)

          ' draw the face
          fg_3Dpolygonobject ColumnData(1,I), 4
        NEXT I
      END IF
    NEXT Col
  NEXT Row

  ' display the scene
  fg_vbpaste 0, %vbWidth-1, 0, %vbHeight-1, 0, %vbHeight-1

  ' display the 3D information at the bottom of the window
  CALL UpdateInfo

END SUB

'*****************************************************************************
'                                                                            *
'  UpdateInfo()                                                              *
'                                                                            *
'  Displays the information at the bottom of the window.                     *
'                                                                            *
'*****************************************************************************

SUB UpdateInfo()

  LOCAL x AS DOUBLE, y AS DOUBLE, Z AS DOUBLE
  LOCAL xDir AS DOUBLE, yDir AS DOUBLE, zDir AS DOUBLE
  LOCAL MessageText AS ASCIIZ*80

  ' get current position and direction
  fg_3Dgetpov x, y, Z, xDir, yDir, zDir

  ' clear an area to write on
  fg_setcolor fg_maprgb(0,0,140)
  fg_rect 0, 249, 0, %InfoHeight-1

  fg_setcolor fg_maprgb(0,140,0)
  fg_rect 250, %vbWidth-1, 0, %InfoHeight-1

  fg_setcolor -1

  ' print current position and unit vector
  fg_move 20, 32
  MessageText = "x = " + FORMAT$(x, "####.##") + "  xDir = " + FORMAT$(xDir, "####.##")
  fg_print MessageText, LEN(MessageText)

  fg_move 20, 46
  MessageText = "y = " + FORMAT$(y, "####.##") + "  yDir = " + FORMAT$(yDir, "####.##")
  fg_print MessageText, LEN(MessageText)

  fg_move 20, 60
  MessageText = "z = " + FORMAT$(Z, "####.##") + "  zDir = " + FORMAT$(zDir, "####.##")
  fg_print MessageText, LEN(MessageText)

  ' print instructions
  fg_move 270, 18
  MessageText = "Up    = move forward   Home = move up"
  fg_print MessageText, LEN(MessageText)

  fg_move 270, 32
  MessageText = "Down  = move back      End  = move down"
  fg_print MessageText, LEN(MessageText)

  fg_move 270, 46
  MessageText = "Left  = turn left      PgUp = look up"
  fg_print MessageText, LEN(MessageText)

  fg_move 270, 60
  MessageText = "Right = turn right     PgDn = look down"
  fg_print MessageText, LEN(MessageText)

  fg_move 290, 74
   MessageText = "Shift+Left/Right = move left/right"
  fg_print MessageText, LEN(MessageText)

  fg_vbpaste 0, %vbWidth-1, 0, %InfoHeight-1, 0, %WinHeight-1

END SUB

'*****************************************************************************
'                                                                            *
'  SetWindowSize()                                                           *
'                                                                            *
'  Sets the window size so the client area has the specified dimensions.     *
'                                                                            *
'*****************************************************************************

SUB SetWindowSize(ClientWidth AS LONG, ClientHeight AS LONG)

  LOCAL ClientRect AS RECT
  LOCAL WindowRect AS RECT
  LOCAL WindowWidth AS LONG, WindowHeight AS LONG

  GetClientRect GetActiveWindow(), ClientRect
  GetWindowRect GetActiveWindow(), WindowRect

  WindowWidth = ClientWidth _
    + (WindowRect.nRight - WindowRect.nLeft) _
    - (ClientRect.nRight - ClientRect.nLeft)
  WindowHeight = ClientHeight _
    + (WindowRect.nBottom - WindowRect.nTop) _
    - (ClientRect.nBottom - ClientRect.nTop)
  SetWindowPos GetActiveWindow(), %NULL, 0, 0, _
     WindowWidth, WindowHeight, %SWP_NOMOVE OR %SWP_NOZORDER

END SUB

'*****************************************************************************
'                                                                            *
'  InitGlobals()                                                             *
'                                                                            *
'  Initialize global variables and arrays. Called from the WM_CREATE handler.*
'                                                                            *
'*****************************************************************************

SUB InitGlobals()

  ' top
  ColumnData(1,1) = -1.0: ColumnData(2,1) = 10.0: ColumnData(3,1) = 1.0
  ColumnData(4,1) = 1.0: ColumnData(5,1) = 10.0: ColumnData(6,1) = 1.0
  ColumnData(7,1) = 1.0: ColumnData(8,1) = 10.0: ColumnData(9,1) = -1.0
  ColumnData(10,1) = -1.0: ColumnData(11,1) = 10.0: ColumnData(12,1) = -1.0

  ' front
  ColumnData(1,2) = -1.0: ColumnData(2,2) = 10.0: ColumnData(3,2) = -1.0
  ColumnData(4,2) = 1.0: ColumnData(5,2) = 10.0: ColumnData(6,2) = -1.0
  ColumnData(7,2) = 1.0: ColumnData(8,2) = 0.0: ColumnData(9,2) = -1.0
  ColumnData(10,2) = -1.0: ColumnData(11,2) = 0.0: ColumnData(12,2) = -1.0

  ' left
  ColumnData(1,3) = -1.0: ColumnData(2,3) = 10.0: ColumnData(3,3) = 1.0
  ColumnData(4,3) = -1.0: ColumnData(5,3) = 10.0: ColumnData(6,3) = -1.0
  ColumnData(7,3) = -1.0: ColumnData(8,3) = 0.0: ColumnData(9,3) = -1.0
  ColumnData(10,3) = -1.0: ColumnData(11,3) = 0.0: ColumnData(12,3) = 1.0

  ' right
  ColumnData(1,4) = 1.0: ColumnData(2,4) = 10.0: ColumnData(3,4) = -1.0
  ColumnData(4,4) = 1.0: ColumnData(5,4) = 10.0: ColumnData(6,4) = 1.0
  ColumnData(7,4) = 1.0: ColumnData(8,4) = 0.0: ColumnData(9,4) = 1.0
  ColumnData(10,4) = 1.0: ColumnData(11,4) = 0.0: ColumnData(12,4) = -1.0

  ' bottom
  ColumnData(1,5) = -1.0: ColumnData(2,5) = 0.0: ColumnData(3,5) = -1.0
  ColumnData(4,5) = 1.0: ColumnData(5,5) = 0.0: ColumnData(6,5) = -1.0
  ColumnData(7,5) = 1.0: ColumnData(8,5) = 0.0: ColumnData(9,5) = 1.0
  ColumnData(10,5) = -1.0: ColumnData(11,5) = 0.0: ColumnData(12,5) = 1.0

  ' back
  ColumnData(1,6) = 1.0: ColumnData(2,6) = 10.0: ColumnData(3,6) = 1.0
  ColumnData(4,6) = -1.0: ColumnData(5,6) = 10.0: ColumnData(6,6) = 1.0
  ColumnData(7,6) = -1.0: ColumnData(8,6) = 0.0: ColumnData(9,6) = 1.0
  ColumnData(10,6) = 1.0: ColumnData(11,6) = 0.0: ColumnData(12,6) = 1.0

END SUB
