'*****************************************************************************
'                                                                            *
'  Fade.bas                                                                  *
'                                                                            *
'  This program shows how perform a palette fade.                            *
'                                                                            *
'*****************************************************************************

#COMPILE EXE "Fade.exe"

#INCLUDE "FGWin.inc"
#RESOURCE "Fade.pbr"

%vbWidth = 320
%vbHeight = 200

DECLARE FUNCTION AppPath() AS STRING

GLOBAL Original() AS BYTE
GLOBAL Current() AS BYTE

'------------------------------------------------------------------------------
' Menu identifiers much match those in Fade.h
%IDM_FadeIn  = 1
%IDM_FadeOut = 2
%IDM_Exit    = 3

'------------------------------------------------------------------------------

FUNCTION WINMAIN (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  lpCmdLine           AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

  LOCAL Msg         AS tagMsg
  LOCAL wndclass    AS WndClassEx
  LOCAL szClassName AS ASCIIZ * 80
  LOCAL hWnd        AS LONG

  DIM Original(1:236*3)
  DIM Current(1:236*3)

  szClassName            = "FGfade"

  wndclass.cbSize        = SIZEOF(WndClass)
  wndclass.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_OWNDC
  wndclass.lpfnWndProc   = CODEPTR(WindowProc)
  wndclass.cbClsExtra    = 0
  wndclass.cbWndExtra    = 0
  wndclass.hInstance     = hInstance
  wndclass.hIcon         = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  wndclass.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wndclass.hbrBackground = %NULL
  wndclass.lpszMenuName  = VARPTR(szClassName)
  wndclass.lpszClassName = VARPTR(szClassName)
  wndclass.hIconSm       = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  RegisterClassEx wndclass

  hWnd = CreateWindow( _
    szClassName, _               ' window class name
    "Palette Fades", _           ' window caption
    %WS_OVERLAPPEDWINDOW, _      ' window style
    %CW_USEDEFAULT, _            ' initial x position
    %CW_USEDEFAULT, _            ' initial y position
    %CW_USEDEFAULT, _            ' initial x size
    %CW_USEDEFAULT, _            ' initial y size
    %NULL, _                     ' parent window handle
    %NULL, _                     ' window menu handle
    hInstance, _                 ' program instance handle
    BYVAL %NULL)                 ' creation parameters

  ShowWindow hWnd, iCmdShow
  UpdateWindow hWnd

  WHILE GetMessage(Msg, %NULL, 0, 0)
    TranslateMessage Msg
    DispatchMessage Msg
  WEND

  FUNCTION = Msg.wParam

END FUNCTION

'------------------------------------------------------------------------------

GLOBAL hDC AS LONG
GLOBAL hMenu AS LONG
GLOBAL hPal AS LONG
GLOBAL hVB AS LONG
GLOBAL cxClient AS LONG, cyClient AS LONG

FUNCTION WindowProc (BYVAL hWnd AS LONG, BYVAL iMsg AS LONG, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) EXPORT AS LONG

  LOCAL ps AS PAINTSTRUCT

  SELECT CASE iMsg

    CASE %WM_CREATE
      hDC = GetDC(hWnd)
      fg_setdc hDC
      hPal = fg_defpal()
      fg_realize hPal

      fg_vbinit
      hVB = fg_vballoc(%vbWidth, %vbHeight)
      fg_vbopen hVB
      fg_vbcolors

      fg_showpcx AppPath()+"\MOUSE.PCX", %FG_AT_XY
      fg_getdacs 10, 236, Original(1)

      hMenu = GetMenu(hWnd)
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_COMMAND
      SELECT CASE wParam

        CASE %IDM_FadeIn
          CALL FadeIn_Click

        CASE %IDM_FadeOut
          CALL FadeOut_Click

        CASE %IDM_Exit
          DestroyWindow hWnd
      END SELECT
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_PAINT
      BeginPaint hWnd, ps
      fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
      EndPaint hWnd, ps
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SETFOCUS
      fg_realize hPal
      InvalidateRect hWnd, BYVAL %NULL, %TRUE
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SIZE
      cxClient = LOWRD(lParam)
      cyClient = HIWRD(lParam)
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_DESTROY
      fg_vbclose
      fg_vbfree hVB
      fg_vbfin
      DeleteObject hPal
      ReleaseDC hWnd, hDC
      PostQuitMessage 0
      FUNCTION = 0
      EXIT FUNCTION

  END SELECT

  FUNCTION = DefWindowProc(hWnd, iMsg, wParam, lParam)

END FUNCTION

FUNCTION AppPath() AS STRING
  LOCAL ModulePath AS ASCIIZ * 256
  GetModuleFileName %NULL, ModulePath, SIZEOF(ModulePath)
  FUNCTION = LEFT$(ModulePath, INSTR(-1, ModulePath, "\")-1)
END FUNCTION

'*****************************************************************************
'                                                                            *
'  FadeIn_Click()                                                            *
'                                                                            *
'  Fade an image back to its original colors.                                *
'                                                                            *
'*****************************************************************************

SUB FadeIn_Click()
  LOCAL I AS LONG
  LOCAL Fading AS LONG

  MOUSEPTR 11
  RESET Current()

  DO
    Fading = %FALSE
    FOR I = 1 TO 236*3
      IF Current(I) <> Original(I) THEN
        INCR Current(I)
        Fading = %TRUE
      END IF
    NEXT I
    fg_setdacs 10, 236, Current(1)
    IF fg_colors > 8 THEN
      fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
    END IF
  LOOP WHILE Fading

  EnableMenuItem hMenu, %IDM_FadeIn, %MF_GRAYED
  EnableMenuItem hMenu, %IDM_FadeOut, %MF_ENABLED
  DrawMenuBar GetActiveWindow()
  MOUSEPTR 0
END SUB

'*****************************************************************************
'                                                                            *
'  FadeOut_Click()                                                           *
'                                                                            *
'  Fade an image to black.                                                   *
'                                                                            *
'*****************************************************************************

SUB FadeOut_Click()
  LOCAL I AS LONG
  LOCAL Fading AS LONG

  MOUSEPTR 11
  FOR I = 1 TO 236*3
    Current(I) = Original(I)
  NEXT I

  DO
    Fading = %FALSE
    FOR I = 1 TO 236*3
      IF Current(I) <> 0 THEN
        DECR Current(I)
        Fading = %TRUE
      END IF
    NEXT I
    fg_setdacs 10, 236, Current(1)
    IF fg_colors > 8 THEN
      fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
    END IF
  LOOP WHILE Fading

  EnableMenuItem hMenu, %IDM_FadeOut, %MF_GRAYED
  EnableMenuItem hMenu, %IDM_FadeIn, %MF_ENABLED
  DrawMenuBar GetActiveWindow()
  MOUSEPTR 0
END SUB
