'*****************************************************************************
'                                                                            *
'  Scale.bas                                                                 *
'                                                                            *
'  This program demonstrates the Fastgraph for Windows 256-color bitmap      *
'  scaling and shearing functions.                                           *
'                                                                            *
'*****************************************************************************

#COMPILE EXE "Scale.exe"

#INCLUDE "FGWin.inc"

%vbWidth = 320
%vbHeight = 240

DECLARE FUNCTION AppPath() AS STRING

GLOBAL Bird() AS BYTE
GLOBAL ScaledBird() AS BYTE

'------------------------------------------------------------------------------

FUNCTION WINMAIN (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  lpCmdLine           AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

  LOCAL Msg         AS tagMsg
  LOCAL wndclass    AS WndClassEx
  LOCAL szClassName AS ASCIIZ * 80
  LOCAL hWnd        AS LONG

  DIM Bird(1:40*20)
  DIM ScaledBird(1:80*40)

  szClassName            = "FGscale"

  wndclass.cbSize        = SIZEOF(WndClass)
  wndclass.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_OWNDC
  wndclass.lpfnWndProc   = CODEPTR(WindowProc)
  wndclass.cbClsExtra    = 0
  wndclass.cbWndExtra    = 0
  wndclass.hInstance     = hInstance
  wndclass.hIcon         = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  wndclass.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wndclass.hbrBackground = %NULL
  wndclass.lpszMenuName  = %NULL
  wndclass.lpszClassName = VARPTR(szClassName)
  wndclass.hIconSm       = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  RegisterClassEx wndclass

  hWnd = CreateWindow( _
    szClassName, _               ' window class name
    "Bitmap Scaling and Shearing", _ ' window caption
    %WS_OVERLAPPEDWINDOW, _      ' window style
    %CW_USEDEFAULT, _            ' initial x position
    %CW_USEDEFAULT, _            ' initial y position
    %CW_USEDEFAULT, _            ' initial x size
    %CW_USEDEFAULT, _            ' initial y size
    %NULL, _                     ' parent window handle
    %NULL, _                     ' window menu handle
    hInstance, _                 ' program instance handle
    BYVAL %NULL)                 ' creation parameters

  ShowWindow hWnd, iCmdShow
  UpdateWindow hWnd

  WHILE GetMessage(Msg, %NULL, 0, 0)
    TranslateMessage Msg
    DispatchMessage Msg
  WEND

  FUNCTION = Msg.wParam

END FUNCTION

'------------------------------------------------------------------------------

GLOBAL hDC AS LONG
GLOBAL hPal AS LONG
GLOBAL hVB AS LONG
GLOBAL cxClient AS LONG, cyClient AS LONG

FUNCTION WindowProc (BYVAL hWnd AS LONG, BYVAL iMsg AS LONG, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) EXPORT AS LONG

  LOCAL ps AS PAINTSTRUCT

  SELECT CASE iMsg

    CASE %WM_CREATE
      hDC = GetDC(hWnd)
      fg_setdc hDC
      hPal = fg_defpal()
      fg_realize hPal

      fg_vbinit
      hVB = fg_vballoc(%vbWidth, %vbHeight)
      fg_vbopen hVB
      fg_vbcolors

      fg_setcolor 20
      fg_fillpage
      CALL LoadBird

      ' draw 40x20 bird (original bitmap)
      fg_move 6, 80
      fg_drwimage Bird(1), 40, 20

      ' draw 50x25 bird
      fg_move 48, 80
      fg_scale Bird(1), ScaledBird(1), 40, 20, 50, 25
      fg_drwimage ScaledBird(1), 50, 25

      ' draw 60x30 bird
      fg_move 100, 80
      fg_scale Bird(1), ScaledBird(1), 40, 20, 60, 30
      fg_drwimage ScaledBird(1), 60, 30

      ' draw 70x35 bird
      fg_move 162, 80
      fg_scale Bird(1), ScaledBird(1), 40, 20, 70, 35
      fg_drwimage ScaledBird(1), 70, 35

      ' draw 80x40 bird
      fg_move 234, 80
      fg_scale Bird(1), ScaledBird(1), 40, 20, 80, 40
      fg_drwimage ScaledBird(1), 80, 40

      ' draw 50x20 bird sheared left horizontally
      fg_move 55, 160
      fg_shear Bird(1), ScaledBird(1), 40, 20, 50, 0
      fg_drwimage ScaledBird(1), 50, 20

      ' draw 50x20 bird sheared right horizontally
      fg_move 115, 160
      fg_shear Bird(1), ScaledBird(1), 40, 20, 50, 1
      fg_drwimage ScaledBird(1), 50, 20

      ' draw 40x30 bird sheared left vertically
      fg_move 175, 160
      fg_shear Bird(1), ScaledBird(1), 40, 20, 30, 2
      fg_drwimage ScaledBird(1), 40, 30

      ' draw 40x30 bird sheared left vertically
      fg_move 225, 160
      fg_shear Bird(1), ScaledBird(1), 40, 20, 30, 3
      fg_drwimage ScaledBird(1), 40, 30
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_PAINT
      BeginPaint hWnd, ps
      fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
      EndPaint hWnd, ps
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SETFOCUS
      fg_realize hPal
      InvalidateRect hWnd, BYVAL %NULL, %TRUE
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SIZE
      cxClient = LOWRD(lParam)
      cyClient = HIWRD(lParam)
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_DESTROY
      fg_vbclose
      fg_vbfree hVB
      fg_vbfin
      DeleteObject hPal
      ReleaseDC hWnd, hDC
      PostQuitMessage 0
      FUNCTION = 0
      EXIT FUNCTION

  END SELECT

  FUNCTION = DefWindowProc(hWnd, iMsg, wParam, lParam)

END FUNCTION

FUNCTION AppPath() AS STRING
  LOCAL ModulePath AS ASCIIZ * 256
  GetModuleFileName %NULL, ModulePath, SIZEOF(ModulePath)
  FUNCTION = LEFT$(ModulePath, INSTR(-1, ModulePath, "\")-1)
END FUNCTION

SUB LoadBird()
  LOCAL I AS LONG
  OPEN AppPath()+"\Bird256.dat" FOR INPUT ACCESS READ AS #1
  FOR I = 1 TO 40*20
    INPUT #1, Bird(I)
  NEXT I
  CLOSE #1
END SUB
