'*****************************************************************************
'                                                                            *
'  Tunnel.bas                                                                *
'                                                                            *
'  This program draws a Gouraud-shaded tunnel and allows the viewer to move  *
'  through the tunnel using keyboard controls.                               *
'                                                                            *
'*****************************************************************************

#COMPILE EXE "Tunnel.exe"

#INCLUDE "FGWin.inc"

%vbWidth = 640
%vbHeight = 480

TYPE POINT3D
  X AS DOUBLE
  Y AS DOUBLE
  Z AS DOUBLE
END TYPE

' four sides of a 20x40x100 tunnel, defined in 3D world coordinates
GLOBAL Faces() AS POINT3D

' RGB color values at each vertex of each side
GLOBAL FacesRGB() AS BYTE

'------------------------------------------------------------------------------

FUNCTION WINMAIN (BYVAL hInstance     AS LONG, _
                  BYVAL hPrevInstance AS LONG, _
                  lpCmdLine           AS ASCIIZ PTR, _
                  BYVAL iCmdShow      AS LONG) AS LONG

  LOCAL Msg         AS tagMsg
  LOCAL wndclass    AS WndClassEx
  LOCAL szClassName AS ASCIIZ * 80
  LOCAL hWnd        AS LONG

  DIM Faces(1:4,1:4)
  DIM FacesRGB(1:12,1:4)

  szClassName            = "FGtunnel"

  wndclass.cbSize        = SIZEOF(WndClass)
  wndclass.style         = %CS_HREDRAW OR %CS_VREDRAW OR %CS_OWNDC
  wndclass.lpfnWndProc   = CODEPTR(WindowProc)
  wndclass.cbClsExtra    = 0
  wndclass.cbWndExtra    = 0
  wndclass.hInstance     = hInstance
  wndclass.hIcon         = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  wndclass.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wndclass.hbrBackground = %NULL
  wndclass.lpszMenuName  = %NULL
  wndclass.lpszClassName = VARPTR(szClassName)
  wndclass.hIconSm       = LoadIcon(hInstance, BYVAL %IDI_APPLICATION)
  RegisterClassEx wndclass

  hWnd = CreateWindow( _
    szClassName, _               ' window class name
    "Gouraud-Shaded Tunnel", _   ' window caption
    %WS_OVERLAPPEDWINDOW, _      ' window style
    %CW_USEDEFAULT, _            ' initial x position
    %CW_USEDEFAULT, _            ' initial y position
    %CW_USEDEFAULT, _            ' initial x size
    %CW_USEDEFAULT, _            ' initial y size
    %NULL, _                     ' parent window handle
    %NULL, _                     ' window menu handle
    hInstance, _                 ' program instance handle
    BYVAL %NULL)                 ' creation parameters

  ShowWindow hWnd, iCmdShow
  UpdateWindow hWnd

  ' The message loop processes entries placed in the message queue.
  ' When no message is ready, call CheckForMovement() to check if
  ' we want to perform movement.

  WHILE %TRUE
    IF PeekMessage(Msg, %NULL, 0, 0, %PM_REMOVE) THEN
      IF Msg.message = %WM_QUIT THEN
        EXIT LOOP
      ELSE
        TranslateMessage Msg
        DispatchMessage Msg
      END IF
    ELSE
      CALL CheckForMovement
    END IF
  WEND

  FUNCTION = Msg.wParam

END FUNCTION

'------------------------------------------------------------------------------

GLOBAL hDC AS LONG
GLOBAL hPal AS LONG
GLOBAL hVB AS LONG
GLOBAL cxClient AS LONG, cyClient AS LONG

GLOBAL Redraw AS LONG
GLOBAL yAngle AS LONG
GLOBAL zViewer AS DOUBLE

FUNCTION WindowProc (BYVAL hWnd AS LONG, BYVAL iMsg AS LONG, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) EXPORT AS LONG

  LOCAL ps AS PAINTSTRUCT
  LOCAL vbDepth AS LONG

  SELECT CASE iMsg

    CASE %WM_CREATE
      hDC = GetDC(hWnd)
      fg_setdc hDC
      hPal = fg_defpal()
      fg_realize hPal

      fg_vbinit
      vbDepth = fg_colors()
      IF vbDepth < 16 THEN vbDepth = 16
      fg_vbdepth vbDepth
      hVB = fg_vballoc(%vbWidth, %vbHeight)
      fg_vbopen hVB
      fg_vbcolors
      fg_setcolor -1
      fg_fillpage

      fg_3Dviewport 0, %vbWidth-1, 0, %vbHeight-1, 1.0
      fg_3Drenderstate %FG_ZCLIP
      fg_3Dlookat 0.0, 10.0, 50.0, 0.0, 10.0, 100.0
      Redraw = %TRUE

      CALL InitGlobals

      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_PAINT
      BeginPaint hWnd, ps
      fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
      EndPaint hWnd, ps
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SETFOCUS
      fg_realize hPal
      InvalidateRect hWnd, BYVAL %NULL, %TRUE
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_SIZE
      cxClient = LOWRD(lParam)
      cyClient = HIWRD(lParam)
      FUNCTION = 0
      EXIT FUNCTION

    CASE %WM_DESTROY
      fg_vbclose
      fg_vbfree hVB
      fg_vbfin
      DeleteObject hPal
      ReleaseDC hWnd, hDC
      PostQuitMessage 0
      FUNCTION = 0
      EXIT FUNCTION

  END SELECT

  FUNCTION = DefWindowProc(hWnd, iMsg, wParam, lParam)

END FUNCTION

'*****************************************************************************
'                                                                            *
'  CheckForMovement()                                                        *
'                                                                            *
'  The CheckForMovement() function checks for key presses that control the   *
'  cube's movement, and if required redraws the cube at its new position and *
'  orientation. It is called from the WinMain() message loop when there are  *
'  no messages waiting.                                                      *
'                                                                            *
'*****************************************************************************

SUB CheckForMovement()
  ' up arrow moves viewer forward
  IF fg_kbtest(72) = 1 THEN
    CALL fg_3Dmoveforward(2.0)
    Redraw = %TRUE

  ' down arrow moves viewer backward
  ELSEIF fg_kbtest(80) = 1 THEN
    CALL fg_3Dmoveforward(-2.0)
    Redraw = %TRUE

  ' right arrow turns viewer to the right
  ELSEIF fg_kbtest(77) = 1 THEN
    CALL fg_3Drotateright(6*10)
    Redraw = %TRUE

  ' left arrow turns viewer to the left
  ELSEIF fg_kbtest(75) = 1 THEN
    CALL fg_3Drotateright(-6*10)
    Redraw = %TRUE
  END IF

  ' if the viewer's position or orientation changed, redraw the tunnel
  IF Redraw THEN
    ' erase the previous frame from the virtual buffer
    fg_setcolor -1
    fg_fillpage

    ' draw the cube
    CALL DrawTunnel

    ' display what we just drew
    fg_vbscale 0, %vbWidth-1, 0, %vbHeight-1, 0, cxClient-1, 0, cyClient-1
    Redraw = %FALSE
  END IF

END SUB


'*****************************************************************************
'                                                                            *
'  DrawTunnel()                                                              *
'                                                                            *
'  Draws each of the tunnel's four sides in 3D world space.                  *
'                                                                            *
'*****************************************************************************

SUB DrawTunnel()

  LOCAL I AS LONG

  FOR I = 1 TO 4
    fg_3Dshade Faces(1,I).X, FacesRGB(1,I), 4
  NEXT

END SUB

'*****************************************************************************
'                                                                            *
'  InitGlobals()                                                             *
'                                                                            *
'  Initialize global variables and arrays. Called from WM_CREATE.            *
'                                                                            *
'*****************************************************************************

SUB InitGlobals()

  ' Floor
  Faces(1, 1).X = -10.0: Faces(1, 1).Y = 0.0: Faces(1, 1).Z = 100.0
  Faces(2,1).X = -10.0: Faces(2,1).Y = 0.0: Faces(2,1).Z = 200.0
  Faces(3,1).X = 10.0: Faces(3,1).Y = 0.0: Faces(3,1).Z = 200.0
  Faces(4,1).X = 10.0: Faces(4,1).Y = 0.0: Faces(4,1).Z = 100.0
  ' West Side
  Faces(1,2).X = -10.0: Faces(1,2).Y = 0.0: Faces(1,2).Z = 100.0
  Faces(2,2).X = -10.0: Faces(2,2).Y = 40.0: Faces(2,2).Z = 100.0
  Faces(3,2).X = -10.0: Faces(3,2).Y = 40.0: Faces(3,2).Z = 200.0
  Faces(4,2).X = -10.0: Faces(4,2).Y = 0.0: Faces(4,2).Z = 200.0
  ' East Side
  Faces(1,3).X = 10.0: Faces(1,3).Y = 0.0: Faces(1,3).Z = 100.0
  Faces(2,3).X = 10.0: Faces(2,3).Y = 0.0: Faces(2,3).Z = 200.0
  Faces(3,3).X = 10.0: Faces(3,3).Y = 40.0: Faces(3,3).Z = 200.0
  Faces(4,3).X = 10.0: Faces(4,3).Y = 40.0: Faces(4,3).Z = 100.0
  ' Ceiling
  Faces(1,4).X = -10.0: Faces(1,4).Y = 40.0: Faces(1,4).Z = 100.0
  Faces(2,4).X = 10.0: Faces(2,4).Y = 40.0: Faces(2,4).Z = 100.0
  Faces(3,4).X = 10.0: Faces(3,4).Y = 40.0: Faces(3,4).Z = 200.0
  Faces(4,4).X = -10.0: Faces(4,4).Y = 40.0: Faces(4,4).Z = 200.0

  ' Floor RGB
  FacesRGB(1,1) = 192: FacesRGB(2,1) = 192: FacesRGB(3,1) = 192
  FacesRGB(4,1) = 64: FacesRGB(5,1) = 64: FacesRGB(6,1) = 64
  FacesRGB(7,1) = 64: FacesRGB(8,1) = 64: FacesRGB(9,1) = 64
  FacesRGB(10,1) = 192: FacesRGB(11,1) = 192: FacesRGB(12,1) = 192
  ' West Side RGB
  FacesRGB(1,2) = 32: FacesRGB(2,2) = 32: FacesRGB(3,2) = 255
  FacesRGB(4,2) = 32: FacesRGB(5,2) = 32: FacesRGB(6,2) = 255
  FacesRGB(7,2) = 32: FacesRGB(8,2) = 32: FacesRGB(9,2) = 96
  FacesRGB(10,2) = 32: FacesRGB(11,2) = 32: FacesRGB(12,2) = 96
  ' East Side RGB
  FacesRGB(1,3) = 32: FacesRGB(2,3) = 32: FacesRGB(3,3) = 255
  FacesRGB(4,3) = 32: FacesRGB(5,3) = 32: FacesRGB(6,3) = 96
  FacesRGB(7,3) = 32: FacesRGB(8,3) = 32: FacesRGB(9,3) = 96
  FacesRGB(10,3) = 32: FacesRGB(11,3) = 32: FacesRGB(12,3) = 255
  ' Ceiling RGB
  FacesRGB(1,4) = 192: FacesRGB(2,4) = 192: FacesRGB(3,4) = 192
  FacesRGB(4,4) = 192: FacesRGB(5,4) = 192: FacesRGB(6,4) = 192
  FacesRGB(7,4) = 64: FacesRGB(8,4) = 64: FacesRGB(9,4) = 64
  FacesRGB(10,4) = 64: FacesRGB(11,4) = 64: FacesRGB(12,4) = 64

END SUB
